from browser_use.logging_config import setup_logging
setup_logging()
from browser_use.agent.prompts import SystemPrompt
from browser_use.agent.service import Agent
from browser_use.agent.views import ActionModel
from browser_use.agent.views import ActionResult
from browser_use.agent.views import AgentHistoryList
from browser_use.browser.browser import Browser
from browser_use.browser.browser import BrowserConfig
from browser_use.controller.service import Controller
from browser_use.dom.service import DomService

__all__ = [
    'Agent',
    'Browser',
    'BrowserConfig',
    'Controller',
    'DomService',
    'SystemPrompt',
    'ActionResult',
    'ActionModel',
    'AgentHistoryList'
]


import logging
import sys

logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.DEBUG,
    format='%(asctime)s - %(levelname)s - %(message)s',
    handlers=[logging.StreamHandler(sys.stdout)]
)


from typing import Callable, Dict, Type
from pydantic import BaseModel, ConfigDict

class RegisteredAction(BaseModel):
    """Model for a registered action"""
    name: str
    description: str
    function: Callable
    param_model: Type[BaseModel]
    
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    def prompt_description(self) -> str:
        """Get a description of the action for the prompt"""
        skip_keys = ['title']
        description_text = f'{self.description}: \n'
        description_text += '{' + str(self.name) + ': '
        description_text += str(
            {
                param_name: {field_key: field_value for field_key, field_value in param_details.items() 
                            if field_key not in skip_keys}
                for param_name, param_details in self.param_model.schema()['properties'].items()
            }
        )
        description_text += '}'
        return description_text


class ActionModel(BaseModel):
    """Base model for dynamically created action models"""
    # This class will have all the registered actions dynamically added
    
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    def get_index(self) -> int | None:
        """Get the index of the action"""
        # Looks for an 'index' value in any of the action parameters
        # Example data shape: {'clicked_element': {'index': 5}}
        parameters = self.model_dump(exclude_unset=True).values()
        if not parameters:
            return None
        
        for param in parameters:
            if param is not None and 'index' in param:
                return param['index']
        return None
    
    def set_index(self, index: int):
        """Overwrite the index of the action"""
        # Get the action name and params
        action_data = self.model_dump(exclude_unset=True)
        action_name = next(iter(action_data.keys()))
        action_params = getattr(self, action_name)
        
        # Update the index directly on the model
        if hasattr(action_params, 'index'):
            action_params.index = index


class ActionRegistry(BaseModel):
    """Model representing the action registry"""
    actions: Dict[str, RegisteredAction] = {}
    
    def get_prompt_description(self) -> str:
        """Get a description of all actions for the prompt"""
        return '\n'.join([action.prompt_description() for action in self.actions.values()])


import asyncio
import logging
from typing import Callable, Dict, Generic, Optional, Type, TypeVar

from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.prompts import PromptTemplate
from pydantic import BaseModel
import json
import lmnr
from lmnr import Laminar, observe

from browser_use.agent.views import ActionModel, ActionResult
from browser_use.browser.context import BrowserContext
from browser_use.controller.registry.service import Registry
from browser_use.controller.views import (
    ClickElementAction,
    DoneAction,
    GoToUrlAction,
    InputTextAction,
    NoParamsAction,
    OpenTabAction,
    GetAllTabsAction,
    ScrollAction,
    SearchBaiduAction,
    SendKeysAction,
    SwitchTabAction,
    ExtractPageContentAction,
    ClickByPositionAction
)
from browser_use.utils import time_execution_async, time_execution_sync

logger = logging.getLogger(__name__)

Context = TypeVar('Context')


class Controller(Generic[Context]):
    def __init__(
        self,
        exclude_actions: list[str] = [],
        output_model: Optional[Type[BaseModel]] = None,
    ):
        self.exclude_actions = exclude_actions
        self.output_model = output_model
        # Fix: Remove type parameter since Registry is not a generic class
        self.registry = Registry(exclude_actions)
        self._register_default_actions()

    def _register_default_actions(self):
        """Register all default browser actions"""
        
        if self.output_model is not None:
            # Create a new model that extends the output model with success parameter
            class ExtendedOutputModel(self.output_model):  # type: ignore
                success: bool = True

            @self.registry.action(
                'Complete task - with return text and if the task is finished (success=True) or not yet completly finished (success=False), because last step is reached',
                param_model=ExtendedOutputModel,
            )
            async def done(params: ExtendedOutputModel):
                # Exclude success from the output JSON since it's an internal parameter
                output_dict = params.model_dump(exclude={'success'})
                return ActionResult(is_done=True, success=params.success, extracted_content=json.dumps(output_dict))
        else:
            @self.registry.action(
                'Complete task - with return text and if the task is finished (success=True) or not yet completly finished (success=False), because last step is reached',
                param_model=DoneAction,
            )
            async def done(params: DoneAction):
                return ActionResult(is_done=True, extracted_content=params.text)

        # Basic Navigation Actions
        @self.registry.action(
            'Search the query in Baidu in the current tab, the query should be a search query like humans search in Baidu, concrete and not vague or super long. More the single most important items. ',
            param_model=SearchBaiduAction,
        )
        async def search_baidu(params: SearchBaiduAction, browser: BrowserContext):
            page = await browser.get_current_page()
            await page.goto(f'https://www.baidu.com/s?wd={params.query}')
            await page.wait_for_load_state()
            msg = f'🔍  Searched for "{params.query}" in Baidu'
            logger.info(msg)
            return ActionResult(extracted_content=msg, include_in_memory=True)

        @self.registry.action('Navigate to URL in the current tab', param_model=GoToUrlAction)
        async def go_to_url(params: GoToUrlAction, browser: BrowserContext):
            page = await browser.get_current_page()
            await page.goto(params.url)
            await page.wait_for_load_state()
            msg = f'🔗  Navigated to {params.url}'
            logger.info(msg)
            return ActionResult(extracted_content=msg, include_in_memory=True)

        @self.registry.action('Go back', param_model=NoParamsAction)
        async def go_back(_: NoParamsAction, browser: BrowserContext):
            await browser.go_back()
            msg = '🔙  Navigated back'
            logger.info(msg)
            return ActionResult(extracted_content=msg, include_in_memory=True)

        # wait for x seconds
        @self.registry.action('Wait for x seconds default 3')
        async def wait(seconds: int = 3):
            msg = f'🕒  Waiting for {seconds} seconds'
            logger.info(msg)
            await asyncio.sleep(seconds)
            return ActionResult(extracted_content=msg, include_in_memory=True)

        # Element Interaction Actions
        @self.registry.action('Click element', param_model=ClickElementAction)
        async def click_element(params: ClickElementAction, browser: BrowserContext):
            session = await browser.get_session()

            if params.index not in await browser.get_selector_map():
                raise Exception(f'Element with index {params.index} does not exist - retry or use alternative actions')

            element_node = await browser.get_dom_element_by_index(params.index)
            initial_pages = len(session.context.pages)
            logger.info(f'Initial pages count: {initial_pages}')

            # if element has file uploader then dont click
            if await browser.is_file_uploader(element_node):
                msg = f'Index {params.index} - has an element which opens file upload dialog. To upload files please use a specific function to upload files '
                logger.info(msg)
                return ActionResult(extracted_content=msg, include_in_memory=True)

            msg = None

            try:
                download_path = await browser._click_element_node(element_node)
                
                # # Wait for potential new page events to be processed
                # # Check multiple times over a longer period
                # max_checks = 10
                # check_interval = 0.05  # 50ms intervals
                
                # for i in range(max_checks):
                #     await asyncio.sleep(check_interval)
                #     current_pages = len(session.context.pages)
                #     if current_pages > initial_pages:
                #         logger.info(f'New page detected after {(i+1) * check_interval:.2f}s, pages: {initial_pages} -> {current_pages}')
                #         break
                
                if download_path:
                    msg = f'💾  Downloaded file to {download_path}'
                else:
                    msg = f'🖱️  Clicked button with index {params.index}: {element_node.get_all_text_till_next_clickable_element(max_depth=2)}'

                logger.info(msg)
                logger.debug(f'Element xpath: {element_node.xpath}')
                
                # Final check for new pages
                final_pages = len(session.context.pages)
                logger.info(f'Final pages count: {final_pages}')
                if final_pages > initial_pages:
                    new_tab_msg = 'New tab opened - switching to it'
                    msg += f' - {new_tab_msg}'
                    logger.info(new_tab_msg)
                    await browser.switch_to_tab(-1)
                
                return ActionResult(extracted_content=msg, include_in_memory=True)
            except Exception as e:
                logger.warning(f'Element not clickable with index {params.index} - most likely the page changed')
                return ActionResult(error=str(e))

        @self.registry.action('Click by position', param_model=ClickByPositionAction)
        async def click_by_position(params: ClickByPositionAction, browser: BrowserContext):
            page = await browser.get_current_page()
            # await page.mouse.move(params.x, params.y)
            await page.mouse.click(x=params.x, y=params.y)
            msg = f'🖱️  Clicked at position {params.x}, {params.y}'
            logger.info(msg)
            return ActionResult(extracted_content=msg, include_in_memory=True)
        @self.registry.action(
            'Input text into a input interactive element',
            param_model=InputTextAction,
        )
        async def input_text(params: InputTextAction, browser: BrowserContext, has_sensitive_data: bool = False):
            if params.index not in await browser.get_selector_map():
                raise Exception(f'Element index {params.index} does not exist - retry or use alternative actions')

            element_node = await browser.get_dom_element_by_index(params.index)
            await browser._input_text_element_node(element_node, params.text)
            if not has_sensitive_data:
                msg = f'⌨️  Input {params.text} into index {params.index}'
            else:
                msg = f'⌨️  Input sensitive data into index {params.index}'
            logger.info(msg)
            logger.debug(f'Element xpath: {element_node.xpath}')
            return ActionResult(extracted_content=msg, include_in_memory=True)

        # Tab Management Actions
        @self.registry.action('Switch tab', param_model=SwitchTabAction)
        async def switch_tab(params: SwitchTabAction, browser: BrowserContext):
            await browser.switch_to_tab(params.page_id)
            # Wait for tab to be ready
            page = await browser.get_current_page()
            await page.wait_for_load_state()
            msg = f'🔄  Switched to tab {params.page_id}'
            logger.info(msg)
            return ActionResult(extracted_content=msg, include_in_memory=True)

        @self.registry.action('Open url in new tab', param_model=OpenTabAction)
        async def open_tab(params: OpenTabAction, browser: BrowserContext):
            await browser.create_new_tab(params.url)
            msg = f'🔗  Opened new tab with {params.url}'
            logger.info(msg)
            return ActionResult(extracted_content=msg, include_in_memory=True)
        
        @self.registry.action('Get all tabs', param_model=GetAllTabsAction)
        async def get_all_tabs(params: GetAllTabsAction, browser: BrowserContext):
            list = browser.current_state.tabs
            logger.info(list)
            msg = f'🔗  All tabs: {list}'
            logger.info(msg)
            return ActionResult(extracted_content=msg, include_in_memory=True, include_elements=False)

        # Content Actions
        @self.registry.action(
            'Extract page content to retrieve specific information from the page, e.g. all company names, a specifc description, all information about, links with companies in structured format or simply links',
            param_model=ExtractPageContentAction,
        )
        async def extract_content(params: ExtractPageContentAction, browser: BrowserContext):
            page = await browser.get_current_page()
            import markdownify
            # import json
            
            logger.info('Extracting content---')

            content = markdownify.markdownify(await page.content())
            
            

            # prompt = 'Your task is to extract the content of the page. goal: {goal}, Page: {page}'
            # template = PromptTemplate(input_variables=['goal', 'page'], template=prompt)
            try:
                # output = page_extraction_llm.invoke(template.format(goal=params.goal, page=content))
                msg = f'📄  Extracted from page\n: {page.url}\n'
                logger.info(msg)
                return ActionResult(extracted_content=msg, include_in_memory=True)
            except Exception as e:
                logger.debug(f'Error extracting content: {e}')
                msg = f'📄  Extracted from page\n: {content}\n'
                logger.info(msg)
                return ActionResult(extracted_content=content)

        @self.registry.action(
            'Scroll down the page by pixel amount - if no amount is specified, scroll down one page',
            param_model=ScrollAction,
        )
        async def scroll_down(params: ScrollAction, browser: BrowserContext):
            page = await browser.get_current_page()
            if params.amount is not None:
                await page.evaluate(f'window.scrollBy(0, {params.amount});')
            else:
                await page.evaluate('window.scrollBy(0, window.innerHeight);')

            amount = f'{params.amount} pixels' if params.amount is not None else 'one page'
            msg = f'🔍  Scrolled down the page by {amount}'
            logger.info(msg)
            return ActionResult(
                extracted_content=msg,
                include_in_memory=True,
            )

        # scroll up
        @self.registry.action(
            'Scroll up the page by pixel amount - if no amount is specified, scroll up one page',
            param_model=ScrollAction,
        )
        async def scroll_up(params: ScrollAction, browser: BrowserContext):
            page = await browser.get_current_page()
            if params.amount is not None:
                await page.evaluate(f'window.scrollBy(0, -{params.amount});')
            else:
                await page.evaluate('window.scrollBy(0, -window.innerHeight);')

            amount = f'{params.amount} pixels' if params.amount is not None else 'one page'
            msg = f'🔍  Scrolled up the page by {amount}'
            logger.info(msg)
            return ActionResult(
                extracted_content=msg,
                include_in_memory=True,
            )

        @self.registry.action(
            'Scroll element to bottom by index',
            param_model=ScrollAction,
        )
        async def scroll_element_to_bottom_by_index(params: ScrollAction, browser: BrowserContext):
            if params.index not in await browser.get_selector_map():
                raise Exception(f'Element with index {params.index} does not exist - retry or use alternative actions')

            element_node = await browser.get_dom_element_by_index(params.index)
            page = await browser.get_current_page()
            try:
                # 首先确保元素在视图中
                locator = page.locator(f'xpath={element_node.xpath}')
                await locator.scroll_into_view_if_needed()
                
                # 执行JavaScript将元素滚动到底部
                await page.evaluate("""
                (xpath) => {
                    const element = document.evaluate(xpath, document, null, 
                                       XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
                    if (element && element.scrollHeight) {
                        element.scrollTop = element.scrollHeight;
                    }
                }
                """, element_node.xpath)
                
                msg = f'🔍  Scrolled element with index {params.index} to bottom: {element_node.get_all_text_till_next_clickable_element(max_depth=2)}'
                logger.info(msg)
                return ActionResult(extracted_content=msg, include_in_memory=True)
            except Exception as e:
                logger.warning(f'Failed to scroll element with index {params.index} to bottom: {str(e)}')
                return ActionResult(error=str(e))

        # send keys
        @self.registry.action(
            'Send strings of special keys like Escape,Backspace, Insert, PageDown, Delete, Enter, Shortcuts such as `Control+o`, `Control+Shift+T` are supported as well. This gets used in keyboard.press.',
            param_model=SendKeysAction,
        )
        async def send_keys(params: SendKeysAction, browser: BrowserContext):
            page = await browser.get_current_page()

            try:
                await page.keyboard.press(params.keys)
            except Exception as e:
                if 'Unknown key' in str(e):
                    # loop over the keys and try to send each one
                    for key in params.keys:
                        try:
                            await page.keyboard.press(key)
                        except Exception as e:
                            logger.debug(f'Error sending key {key}: {str(e)}')
                            raise e
                else:
                    raise e
            msg = f'⌨️  Sent keys: {params.keys}'
            logger.info(msg)
            return ActionResult(extracted_content=msg, include_in_memory=True)

        @self.registry.action(
            description='If you dont find something which you want to interact with, scroll to it',
        )
        async def scroll_to_text(text: str, browser: BrowserContext):
            page = await browser.get_current_page()
            try:
                # Try different locator strategies
                locators = [
                    page.get_by_text(text, exact=False),
                    page.locator(f'text={text}'),
                    page.locator(f"//*[contains(text(), '{text}')]"),
                ]

                for locator in locators:
                    try:
                        # First check if element exists and is visible
                        if await locator.count() > 0 and await locator.first.is_visible():
                            await locator.first.scroll_into_view_if_needed()
                            await asyncio.sleep(0.5)  # Wait for scroll to complete
                            msg = f'🔍  Scrolled to text: {text}'
                            logger.info(msg)
                            return ActionResult(extracted_content=msg, include_in_memory=True)
                    except Exception as e:
                        logger.debug(f'Locator attempt failed: {str(e)}')
                        continue

                msg = f"Text '{text}' not found or not visible on page"
                logger.info(msg)
                return ActionResult(extracted_content=msg, include_in_memory=True)

            except Exception as e:
                msg = f"Failed to scroll to text '{text}': {str(e)}"
                logger.error(msg)
                return ActionResult(error=msg, include_in_memory=True)

        @self.registry.action(
            description='Get all options from a native dropdown',
        )
        async def get_dropdown_options(index: int, browser: BrowserContext) -> ActionResult:
            """Get all options from a native dropdown"""
            page = await browser.get_current_page()
            selector_map = await browser.get_selector_map()
            dom_element = selector_map[index]

            try:
                # Frame-aware approach since we know it works
                all_options = []
                frame_index = 0

                for frame in page.frames:
                    try:
                        options = await frame.evaluate(
                            """
                            (xpath) => {
                                const select = document.evaluate(xpath, document, null,
                                    XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
                                if (!select) return null;

                                return {
                                    options: Array.from(select.options).map(opt => ({
                                        text: opt.text,
                                        value: opt.value,
                                        index: opt.index
                                    })),
                                    id: select.id,
                                    name: select.name
                                };
                            }
                            """,
                            dom_element.xpath,
                        )

                        if options:
                            logger.debug(f'Found dropdown in frame {frame_index}')
                            logger.debug(f'Dropdown ID: {options["id"]}, Name: {options["name"]}')

                            formatted_options = []
                            for opt in options['options']:
                                # encoding ensures AI uses the exact string in select_dropdown_option
                                encoded_text = json.dumps(opt['text'])
                                formatted_options.append(f'{opt["index"]}: text={encoded_text}')

                            all_options.extend(formatted_options)

                    except Exception as frame_e:
                        logger.debug(f'Frame {frame_index} evaluation failed: {str(frame_e)}')

                    frame_index += 1

                if all_options:
                    msg = '\n'.join(all_options)
                    msg += '\nUse the exact text string in select_dropdown_option'
                    logger.info(msg)
                    return ActionResult(extracted_content=msg, include_in_memory=True)
                else:
                    msg = 'No options found in any frame for dropdown'
                    logger.info(msg)
                    return ActionResult(extracted_content=msg, include_in_memory=True)

            except Exception as e:
                logger.error(f'Failed to get dropdown options: {str(e)}')
                msg = f'Error getting options: {str(e)}'
                logger.info(msg)
                return ActionResult(extracted_content=msg, include_in_memory=True)

        @self.registry.action(
            description='Select dropdown option for interactive element index by the text of the option you want to select',
        )
        async def select_dropdown_option(
            index: int,
            text: str,
            browser: BrowserContext,
        ) -> ActionResult:
            """Select dropdown option by the text of the option you want to select"""
            page = await browser.get_current_page()
            selector_map = await browser.get_selector_map()
            dom_element = selector_map[index]

            # Validate that we're working with a select element
            if dom_element.tag_name != 'select':
                logger.error(f'Element is not a select! Tag: {dom_element.tag_name}, Attributes: {dom_element.attributes}')
                msg = f'Cannot select option: Element with index {index} is a {dom_element.tag_name}, not a select'
                return ActionResult(extracted_content=msg, include_in_memory=True)

            logger.debug(f"Attempting to select '{text}' using xpath: {dom_element.xpath}")
            logger.debug(f'Element attributes: {dom_element.attributes}')
            logger.debug(f'Element tag: {dom_element.tag_name}')

            xpath = '//' + dom_element.xpath

            try:
                frame_index = 0
                for frame in page.frames:
                    try:
                        logger.debug(f'Trying frame {frame_index} URL: {frame.url}')

                        # First verify we can find the dropdown in this frame
                        find_dropdown_js = """
                            (xpath) => {
                                try {
                                    const select = document.evaluate(xpath, document, null,
                                        XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
                                    if (!select) return null;
                                    if (select.tagName.toLowerCase() !== 'select') {
                                        return {
                                            error: `Found element but it's a ${select.tagName}, not a SELECT`,
                                            found: false
                                        };
                                    }
                                    return {
                                        id: select.id,
                                        name: select.name,
                                        found: true,
                                        tagName: select.tagName,
                                        optionCount: select.options.length,
                                        currentValue: select.value,
                                        availableOptions: Array.from(select.options).map(o => o.text.trim())
                                    };
                                } catch (e) {
                                    return {error: e.toString(), found: false};
                                }
                            }
                        """

                        dropdown_info = await frame.evaluate(find_dropdown_js, dom_element.xpath)

                        if dropdown_info:
                            if not dropdown_info.get('found'):
                                logger.error(f'Frame {frame_index} error: {dropdown_info.get("error")}')
                                continue

                            logger.debug(f'Found dropdown in frame {frame_index}: {dropdown_info}')

                            # "label" because we are selecting by text
                            # nth(0) to disable error thrown by strict mode
                            # timeout=1000 because we are already waiting for all network events, therefore ideally we don't need to wait a lot here (default 30s)
                            selected_option_values = (
                                await frame.locator('//' + dom_element.xpath).nth(0).select_option(label=text, timeout=1000)
                            )

                            msg = f'selected option {text} with value {selected_option_values}'
                            logger.info(msg + f' in frame {frame_index}')

                            return ActionResult(extracted_content=msg, include_in_memory=True)

                    except Exception as frame_e:
                        logger.error(f'Frame {frame_index} attempt failed: {str(frame_e)}')
                        logger.error(f'Frame type: {type(frame)}')
                        logger.error(f'Frame URL: {frame.url}')

                    frame_index += 1

                msg = f"Could not select option '{text}' in any frame"
                logger.info(msg)
                return ActionResult(extracted_content=msg, include_in_memory=True)

            except Exception as e:
                msg = f'Selection failed: {str(e)}'
                logger.error(msg)
                return ActionResult(error=msg, include_in_memory=True)

    def action(self, description: str, **kwargs):
        """Decorator for registering custom actions

        @param description: Describe the LLM what the function does (better description == better function calling)
        """
        return self.registry.action(description, **kwargs)

    @time_execution_sync('--act')
    async def act(
        self,
        action: ActionModel,
        browser_context: BrowserContext,
        page_extraction_llm: Optional[BaseChatModel] = None,
        sensitive_data: Optional[Dict[str, str]] = None,
        available_file_paths: Optional[list[str]] = None,
        context: Context | None = None,
    ) -> ActionResult:
        """Execute an action"""

        try:
            for action_name, params in action.model_dump(exclude_unset=True).items():
                if params is not None:
                    result = await self.registry.execute_action(
                        action_name,
                        params,
                        browser=browser_context,
                        page_extraction_llm=page_extraction_llm,
                        sensitive_data=sensitive_data,
                        available_file_paths=available_file_paths,
                        context=context,
                    )

                    if isinstance(result, str):
                        return ActionResult(extracted_content=result)
                    elif isinstance(result, ActionResult):
                        return result
                    elif result is None:
                        return ActionResult()
                    else:
                        raise ValueError(f'Invalid action result type: {type(result)} of {result}')
            return ActionResult()
        except Exception as e:
            raise e

    @observe(name='controller.multi_act')
    @time_execution_async('--multi-act')
    async def multi_act(
        self,
        actions: list[ActionModel],
        browser_context: BrowserContext,
        check_break_if_paused: Callable[[], bool],
        check_for_new_elements: bool = True,
        page_extraction_llm: Optional[BaseChatModel] = None,
        sensitive_data: Optional[Dict[str, str]] = None,
    ) -> list[ActionResult]:
        """Execute multiple actions"""
        results = []
        session = await browser_context.get_session()
        selector_map = session.cached_state.selector_map
        initial_hash_set = set(dom.hash.branch_path_hash for dom in selector_map.values())

        check_break_if_paused()

        for i, action in enumerate(actions):
            logger.info(f'executing action {i}')
            
            check_break_if_paused()
            
            if action.get_index() is not None and i != 0:
                current_state = await browser_context.get_state()
                current_hash_set = set(dom.hash.branch_path_hash for dom in current_state.selector_map.values())
                
                if check_for_new_elements and not current_hash_set.issubset(initial_hash_set):
                    msg = f'Something new appeared after action {i} / {len(actions)}'
                    logger.info(msg)
                    results.append(ActionResult(extracted_content=msg, include_in_memory=True))
                    break
            
            check_break_if_paused()
            
            results.append(await self.act(action, browser_context, page_extraction_llm, sensitive_data))
            
            logger.debug(f'Executed action {i + 1} / {len(actions)}')
            
            if results[-1].is_done or results[-1].error or i == len(actions) - 1:
                break
            
            await asyncio.sleep(browser_context.config.wait_between_actions)
        
        return results


import logging
from datetime import datetime
from importlib import resources
from typing import Optional
from playwright.async_api import Page
from browser_use.dom.history_tree_processor.view import Coordinates
from browser_use.dom.views import CoordinateSet, DOMBaseNode, DOMElementNode, DOMState, DOMTextNode, SelectorMap, ViewportInfo
import json
import os
import sys
from pathlib import Path

logger = logging.getLogger(__name__)

class DomService:
    def __init__(self, page):
        self.page = page
        self.xpath_cache = {}
        self.js_code = resources.read_text('browser_use.dom', 'buildDomTree.js')
        self.recording_json = self._load_recording_json()
        self.elements_config = resources.read_text('browser_use.dom', 'elements-config.json')

    def _load_elements_config(self):
        """加载 elements-config.json 文件，如果文件不存在或无法读取则返回空列表"""
        try:
            config_path = os.path.join(os.getcwd(), 'elements-config.json')
            if os.path.exists(config_path):
                with open(config_path, 'r', encoding='utf-8') as f:
                    return json.load(f)
            else:
                logger.info(f'elements-config.json 文件不存在，使用空列表作为默认值')
                return []
        except Exception as e:
            logger.warning(f'读取 elements-config.json 失败: {e}')
            return []

    def _load_recording_json(self):
        # 获取 cache_dir 路径
        if sys.platform.startswith('linux'):
            cache_dir = '/home/ubuntu/workspace/.cache'
        elif sys.platform == 'darwin':
            cache_dir = os.path.join(Path.home(), 'Library', 'Caches')
        elif sys.platform == 'win32':
            cache_dir = os.environ.get('LOCALAPPDATA', os.path.join(Path.home(), 'AppData', 'Local'))
        else:
            raise RuntimeError(f'Unsupported platform: {sys.platform}')
        # 优先读取 cache_dir 下 orion-recording/recording_task.json
        recording_path = os.path.join(cache_dir, 'orion-recording', 'recording_task.json')
        
        os.makedirs(os.path.join(cache_dir, 'orion-recording'), exist_ok=True)
        print(f'recording_path: {recording_path}')
        if not os.path.exists(recording_path):
                return None
        try:
            with open(recording_path, 'r', encoding='utf-8') as f:
                return json.load(f)
        except Exception as e:
            logger.warning(f'Failed to load recording_task.json: {e}')
            return None

    async def get_clickable_elements(self, highlight_elements=True, focus_element=-1, viewport_expansion=0):
        element_tree, selector_map = await self._build_dom_tree(highlight_elements, focus_element, viewport_expansion)
        
        return DOMState(element_tree=element_tree, selector_map=selector_map)
    
    async def _build_dom_tree(self, highlight_elements, focus_element, viewport_expansion):
        if await self.page.evaluate('1+1') != 2:
            raise ValueError('The page cannot evaluate javascript code properly')
        
        debug_mode = logger.getEffectiveLevel() == logging.DEBUG
        args = {
            'doHighlightElements': highlight_elements,
            'focusHighlightIndex': focus_element,
            'viewportExpansion': viewport_expansion,
            'debugMode': debug_mode,
            'recordingData': self.recording_json,
            'elementsConfig': json.loads(self.elements_config)
        }
        
        try:
            eval_page = await self.page.evaluate(self.js_code, args)
        except Exception as e:
            logger.error('Error evaluating JavaScript: %s', e)
            raise
            
        if debug_mode and 'perfMetrics' in eval_page:
            logger.debug('DOM Tree Building Performance Metrics:\n%s', json.dumps(eval_page['perfMetrics'], indent=2))
            
        return await self._construct_dom_tree(eval_page)
    
    async def _construct_dom_tree(self, eval_page):
        js_node_map = eval_page['map']
        js_root_id = eval_page['rootId']
        
        selector_map = {}
        node_map = {}
        
        for id, node_data in js_node_map.items():
            node, children_ids = self._parse_node(node_data)
            if node is None:
                continue
                
            node_map[id] = node
            
            if isinstance(node, DOMElementNode) and node.highlight_index is not None:
                selector_map[node.highlight_index] = node
                
            if isinstance(node, DOMElementNode):
                for child_id in children_ids:
                    if child_id not in node_map:
                        continue
                        
                    child_node = node_map[child_id]
                    
                    child_node.parent = node
                    node.children.append(child_node)
                    
        html_to_dict = node_map[str(js_root_id)]
        
        if html_to_dict is None or not isinstance(html_to_dict, DOMElementNode):
            raise ValueError('Failed to parse HTML to dictionary')
            
        return html_to_dict, selector_map
    
    def _create_selector_map(self, element_tree):
        selector_map = {}
        
        def traverse_node(node):
            if isinstance(node, DOMElementNode):
                if node.highlight_index is not None:
                    selector_map[node.highlight_index] = node
                for child in node.children:
                    traverse_node(child)
        
        traverse_node(element_tree)
        return selector_map
    
    def _parse_node(self, node_data, viewport=None, parent=None):
        if not node_data:
            return None, []
            
        # Process text nodes immediately
        if node_data.get('type') == 'TEXT_NODE':
            text_node = DOMTextNode(
                text=node_data['text'],
                is_visible=node_data['isVisible'],
                parent=parent
            )
            return text_node, []
            
        tag_name = node_data['tagName']
        viewport_coordinates = None
        page_coordinates = None
        
        if viewport:
            scroll_x = viewport.scroll_x
            scroll_y = viewport.scroll_y
            
            if 'viewportPos' in node_data:
                x1, y1, x2, y2 = node_data['viewportPos']
                x1 = int(x1)
                y1 = int(y1)
                x2 = int(x2)
                y2 = int(y2)
                width = int(x2 - x1)
                height = int(y2 - y1)
                
                viewport_coordinates = CoordinateSet(
                    top_left=Coordinates(x=x1, y=y1),
                    top_right=Coordinates(x=x2, y=y1),
                    bottom_left=Coordinates(x=x1, y=y2),
                    bottom_right=Coordinates(x=x2, y=y2),
                    center=Coordinates(x=int(x1 + width / 2), y=int(y1 + height / 2)),
                    width=width,
                    height=height
                )
                
                page_coordinates = CoordinateSet(
                    top_left=Coordinates(x=x1 + scroll_x, y=y1 + scroll_y),
                    top_right=Coordinates(x=x2 + scroll_x, y=y1 + scroll_y),
                    bottom_left=Coordinates(x=x1 + scroll_x, y=y2 + scroll_y),
                    bottom_right=Coordinates(x=x2 + scroll_x, y=y2 + scroll_y),
                    center=Coordinates(x=int(x1 + scroll_x + width / 2), y=int(y1 + scroll_y + height / 2)),
                    width=width,
                    height=height
                )
                
        element_node = DOMElementNode(
            tag_name=tag_name,
            xpath=node_data['xpath'],
            attributes=node_data.get('attributes', {}),
            children=[],
            is_visible=node_data.get('isVisible', False),
            is_interactive=node_data.get('isInteractive', False),
            is_scroll_element=node_data.get('isScrollElement', False),
            is_top_element=node_data.get('isTopElement', False),
            desc=node_data.get('desc', None),
            highlight_index=node_data.get('highlightIndex'),
            shadow_root=node_data.get('shadowRoot', False),
            parent=parent,
            viewport_coordinates=viewport_coordinates,
            page_coordinates=page_coordinates,
            viewport_info=viewport
        )
        
        children_ids = node_data.get('children', [])
        
        return element_node, children_ids


from typing import Any, List
import bashlex
import bashlex.errors
from app.logger import logger

__all__ = [
    'process_terminal_output',
    'split_bash_commands'
]

def split_bash_commands(commands):
    """
    能够将类似 'ls -l \n echo hello' 这样的命令拆分成两个单独的命令 
    但是类似 'echo a && echo b' 这样的命令不会被拆分 
    Copy from OpenHands:
    https://github.com/All-Hands-AI/OpenHands/blob/main/openhands/runtime/utils/bash.py
    """
    if not commands.strip():
        return ['']
    
    try:
        parsed = bashlex.parse(commands)
        result = []
        start = 0
        
        for node in parsed:
            if hasattr(node, 'pos') and hasattr(node, 'end'):
                cmd = commands[node.pos:node.end].strip()
                result.append(cmd)
        
        if not result and commands.strip():
            if '\n' in commands:
                return [cmd.strip() for cmd in commands.split('\n') if cmd.strip()]
            else:
                return [commands.strip()]
        
        return result
    except (bashlex.errors.ParsingError, Exception) as e:
        if '\n' in commands:
            return [cmd.strip() for cmd in commands.split('\n') if cmd.strip()]
        return [commands.strip()] if commands.strip() else ['']

def process_terminal_output(text):
    '''
    处理终端输出，保留 ANSI 转义序列并正确处理行覆盖
    处理规则：
    1. 保留所有 ANSI 转义序列（\x1b[...m 颜色，\x1b[...G 光标移动等）
    2. 处理 \r 的行内覆盖效果
    3. 处理光标控制序列的行内覆盖效果
    '''
    if not text:
        return ""
    
    lines = text.split('\n')
    result = []
    
    for line in lines:
        if '\r' in line:
            parts = line.split('\r')
            processed_line = parts[0]
            
            for i in range(len(parts) - 1):
                ansi_colors = extract_ansi_colors(parts[i])
                if ansi_colors and not have_matching_ansi_reset(parts[0]):
                    processed_line = ansi_colors + processed_line
            
            result.append(processed_line)
        else:
            processed_line = process_cursor_movements(line)
            result.append(processed_line)
    
    return '\n'.join(result)

def extract_ansi_colors(text):
    """
    Extract ANSI color sequences from a text.
    
    Args:
        text: The text to extract colors from
        
    Returns:
        str: Concatenated color sequences found in the text
    """
    import re
    
    color_pattern = r'\x1b\[\d+(;\d+)*m'
    colors = re.findall(color_pattern, text)
    
    return ''.join(colors)

def have_matching_ansi_reset(text):
    """
    Check if the text has a matching ANSI reset sequence.
    
    Args:
        text: The text to check
        
    Returns:
        bool: True if there's a reset sequence, False otherwise
    """
    return '\x1b[0m' in text or '\x1b[m' in text

def process_cursor_movements(line):
    """
    Process cursor movement escape sequences.
    
    Args:
        line: Line to process
        
    Returns:
        str: Processed line with cursor movements applied
    """
    import re
    
    cursor_pattern = r'\x1b\[(\d+)G'
    matches = list(re.finditer(cursor_pattern, line))
    
    if not matches:
        return line
    
    result = list(line)
    for match in reversed(matches):
        col = int(match.group(1)) - 1
        cmd_start = match.start()
        cmd_end = match.end()
        
        for i in range(cmd_start, cmd_end):
            result[i] = ''
    
    return ''.join(result)


import hashlib
from dataclasses import dataclass # in bytecode not accessed?
from typing import Optional # in bytecode not accessed?
from browser_use.dom.history_tree_processor.view import DOMHistoryElement, HashedDomElement
from browser_use.dom.views import DOMElementNode


class HistoryTreeProcessor:
    """
    Operations on the DOM elements

    @dev be careful - text nodes can change even if elements stay the same
    """

    @staticmethod
    def convert_dom_element_to_history_element(dom_element):
        parent_branch_path = HistoryTreeProcessor._get_parent_branch_path(dom_element)
        css_selector = dom_element.get_advanced_css_selector()
        return DOMHistoryElement(
            dom_element.tag_name,
            dom_element.xpath,
            dom_element.highlight_index,
            parent_branch_path,
            dom_element.attributes,
            dom_element.shadow_root,
            css_selector=css_selector,
            page_coordinates=dom_element.page_coordinates,
            viewport_coordinates=dom_element.viewport_coordinates,
            viewport_info=dom_element.viewport_info
        )

    @staticmethod
    def find_history_element_in_tree(dom_history_element, tree):
        hashed_dom_history_element = HistoryTreeProcessor._hash_dom_history_element(dom_history_element)
        
        def process_node(node):
            if node.highlight_index is not None:
                hashed_node = HistoryTreeProcessor._hash_dom_element(node)
                if hashed_node == hashed_dom_history_element:
                    return node
            for child in node.children:
                if isinstance(child, DOMElementNode):
                    result = process_node(child)
                    if result is not None:
                        return result
            return None
            
        return process_node(tree)

    @staticmethod
    def compare_history_element_and_dom_element(dom_history_element, dom_element):
        hashed_dom_history_element = HistoryTreeProcessor._hash_dom_history_element(dom_history_element)
        hashed_dom_element = HistoryTreeProcessor._hash_dom_element(dom_element)
        return hashed_dom_history_element == hashed_dom_element

    @staticmethod
    def _hash_dom_history_element(dom_history_element):
        branch_path_hash = HistoryTreeProcessor._parent_branch_path_hash(dom_history_element.entire_parent_branch_path)
        attributes_hash = HistoryTreeProcessor._attributes_hash(dom_history_element.attributes)
        xpath_hash = HistoryTreeProcessor._xpath_hash(dom_history_element.xpath)
        return HashedDomElement(branch_path_hash, attributes_hash, xpath_hash)

    @staticmethod
    def _hash_dom_element(dom_element):
        parent_branch_path = HistoryTreeProcessor._get_parent_branch_path(dom_element)
        branch_path_hash = HistoryTreeProcessor._parent_branch_path_hash(parent_branch_path)
        attributes_hash = HistoryTreeProcessor._attributes_hash(dom_element.attributes)
        xpath_hash = HistoryTreeProcessor._xpath_hash(dom_element.xpath)
        return HashedDomElement(branch_path_hash, attributes_hash, xpath_hash)

    @staticmethod
    def _get_parent_branch_path(dom_element):
        parents = []
        current_element = dom_element
        while current_element.parent is not None:
            parents.append(current_element)
            current_element = current_element.parent
            
        parents.reverse()
        
        return [parent.tag_name for parent in parents]

    @staticmethod
    def _parent_branch_path_hash(parent_branch_path):
        parent_branch_path_string = '/'.join(parent_branch_path)
        return hashlib.sha256(parent_branch_path_string.encode()).hexdigest()

    @staticmethod
    def _attributes_hash(attributes):
        attributes_string = ''.join(f'{key}={value}' for key, value in attributes.items())
        return hashlib.sha256(attributes_string.encode()).hexdigest()

    @staticmethod
    def _xpath_hash(xpath):
        return hashlib.sha256(xpath.encode()).hexdigest()

    @staticmethod
    def _text_hash(dom_element):
        text_string = dom_element.get_all_text_till_next_clickable_element()
        return hashlib.sha256(text_string.encode()).hexdigest()


from dataclasses import dataclass, field
from typing import Any, Optional
from playwright.async_api import ElementHandle
from pydantic import BaseModel
from browser_use.dom.history_tree_processor.service import DOMHistoryElement
from browser_use.dom.views import DOMState


class TabInfo(BaseModel):
    """Represents information about a browser tab"""
    page_id: int
    url: str
    title: str


@dataclass
class ClickableElementData:
    element_handle: ElementHandle
    element_info: "ElementInfo"
    x_center: float
    y_center: float


@dataclass
class BrowserState(DOMState):
    url: str
    title: str
    tabs: list[TabInfo]
    screenshot: Optional[str] = None
    pixels_above: int = 0
    pixels_below: int = 0
    browser_errors: list[str] = field(default_factory=list)


@dataclass
class ElementInfo:
    element_id: str
    tag_name: str
    attributes: dict[str, str]
    text_content: str
    inner_text: str
    visible_text: str
    bounding_box: dict[str, float]
    is_visible: bool
    computed_style: dict[str, str]


@dataclass
class ExtractedPageContentInfo:
    elements: list[ElementInfo]
    page_text: str
    html_content: str


@dataclass
class BrowserStateHistory:
    url: str
    title: str
    tabs: list[TabInfo]
    interacted_element: list[DOMHistoryElement | None] | list[None]
    screenshot: Optional[str] = None
    
    def to_dict(self) -> dict[str, Any]:
        data = {}
        data['tabs'] = [tab.model_dump() for tab in self.tabs]
        data['screenshot'] = self.screenshot
        data['interacted_element'] = [el.to_dict() if el else None for el in self.interacted_element]
        data['url'] = self.url
        data['title'] = self.title
        return data


class BrowserError(Exception):
    """Base class for all browser errors"""
    pass


class URLNotAllowedError(BrowserError):
    """Error raised when a URL is not allowed"""
    pass


class ScreenshotError(BrowserError):
    """Error raised when a screenshot fails"""
    pass